/*******************************************************************************
  MPLAB Harmony Application Source File

  Company:
    Microchip Technology Inc.

  File Name:
    app.c

  Summary:
    This file contains the source code for the MPLAB Harmony application.

  Description:
    This file contains the source code for the MPLAB Harmony application.  It
    implements the logic of the application's state machine and it may call
    API routines of other MPLAB Harmony modules in the system, such as drivers,
    system services, and middleware.  However, it does not call any of the
    system interfaces (such as the "Initialize" and "Tasks" functions) of any of
    the modules in the system or make any assumptions about when those functions
    are called.  That is the responsibility of the configuration-specific system
    files.
 *******************************************************************************/

// *****************************************************************************
// *****************************************************************************
// Section: Included Files
// *****************************************************************************
// *****************************************************************************

#include "app.h"
#include "cryptoauthlib.h"
#include "../../slot_6_secret_key.h"

// *****************************************************************************
// *****************************************************************************
// Section: Global Data Definitions
// *****************************************************************************
// *****************************************************************************
#define BOOT_START_ADDR                  (0x00000000)
#define APP_START_ADDR                   (0x0000C000)
#define SIGNATURE_START_ADDR             (0x0003FC00)
#define APP_LEN_PRESENT_ADDR             (0x0003FD00)

extern ATCAIfaceCfg atecc608_0_init_data;
extern uint8_t slot_6_secret_key[];
uint8_t digest[ATCA_SHA_DIGEST_SIZE];
uint8_t signature[ATCA_SIG_SIZE];
unsigned int volatile * const app_length = (unsigned int *) APP_LEN_PRESENT_ADDR;

// *****************************************************************************
/* Application Data

  Summary:
    Holds application data

  Description:
    This structure holds the application's data.

  Remarks:
    This structure should be initialized by the APP_Initialize function.

    Application strings and buffers are be defined outside this structure.
*/

APP_DATA appData;

// *****************************************************************************
// *****************************************************************************
// Section: Application Callback Functions
// *****************************************************************************
// *****************************************************************************

/* TODO:  Add any necessary callback functions.
*/

// *****************************************************************************
// *****************************************************************************
// Section: Application Local Functions
// *****************************************************************************
// *****************************************************************************

void RNG(uint8_t* dest, size_t size)
{
    ATCA_STATUS status;
    uint8_t random[RANDOM_NUM_SIZE];
    
    if (ATCA_SUCCESS == (status = atcab_random(random)))
    {
        if (dest && (size <= RANDOM_NUM_SIZE))
        {
            memcpy(dest, random, size);
        }
    }
}

ATCA_STATUS calc_digest_and_get_signature(void)
{
    ATCA_STATUS status = ATCA_GEN_FAIL;
    atcac_sha2_256_ctx firm_ctx;
    char displaystr[400];
    size_t displaylen;

    do
    {
        if (ATCA_SUCCESS != (status = atcac_sw_sha2_256_init(&firm_ctx)))
        {
            printf("atcac_sw_sha2_256_init failed with an error %02X\r\n", status);
            break;
        }

        if (ATCA_SUCCESS != (status = atcac_sw_sha2_256_update(&firm_ctx, BOOT_START_ADDR, 
                                                               APP_START_ADDR + *app_length)))
        {
            printf("atcac_sw_sha2_256_update failed with an error %02X\r\n", status);
            break;
        }

        if (ATCA_SUCCESS != (status = atcac_sw_sha2_256_finish(&firm_ctx, digest)))
        {
            printf("atcac_sw_sha2_256_update failed with an error %02X\r\n", status);
            break;
        }
    } 
    while (0);

    if (ATCA_SUCCESS == status)
    {
        memcpy(signature, (void* )SIGNATURE_START_ADDR, ATCA_SIG_SIZE);

        displaylen = sizeof(displaystr);
        atcab_bin2hex(digest, ATCA_SHA_DIGEST_SIZE, displaystr, &displaylen);
        printf("\r\nCalculated Application Digest:\r\n%s\r\n", displaystr);

        displaylen = sizeof(displaystr);
        atcab_bin2hex(signature, ATCA_SIG_SIZE, displaystr, &displaylen);
        printf("\r\nApplication Siganture from firmware:\r\n%s\r\n", displaystr);
    }
    
    return status;
}

ATCA_STATUS validate_firmware_based_on_stored_digest(void)
{
    ATCA_STATUS status = ATCA_GEN_FAIL;
    uint8_t host_random[RANDOM_NUM_SIZE];
    bool is_verified = false;

    RNG(host_random, RANDOM_NUM_SIZE);

    /* NOTE: In real applications, IO Protection pairing step should be done on the first
             power-up and lock IO protection key (SLOT 6 here) to avoid further updates to it.
    */

    if (ATCA_SUCCESS == (status = atcab_secureboot_mac(SECUREBOOT_MODE_FULL_STORE, digest, signature,
                                                       host_random, slot_6_secret_key, &is_verified)))
    {
        if (is_verified)
        {
            printf("\r\nFirmware is validated successfully based on stored digest on SE\r\n");
        }
        else
        {
            status = !ATCA_SUCCESS;
        }
    }

    return status;
}

ATCA_STATUS validate_firmware_and_update_digest(void)
{
    ATCA_STATUS status = ATCA_GEN_FAIL;
    uint8_t host_random[RANDOM_NUM_SIZE];
    bool is_verified = false;

    RNG(host_random, RANDOM_NUM_SIZE);

    /* NOTE: In real applications, IO Protection pairing step should be done on the first
             power-up and lock IO protection key (SLOT 6 here) to avoid further updates to it.
    */

    if (ATCA_SUCCESS == (status = atcab_secureboot_mac(SECUREBOOT_MODE_FULL_COPY, digest, signature,
                                                       host_random, slot_6_secret_key, &is_verified)))
    {
        if (is_verified)
        {
            printf("\r\nFirmware is validated successfully!!! and Updated new digest into SE\r\n");
        }
        else
        {
            status = !ATCA_SUCCESS;
        }
    }

    if (ATCA_SUCCESS != status)
    {
        printf("\r\nFirmware is not validated successfully");
    }

    return status;
}

void jump_to_application(void)
{
    uint32_t app_start_addr = *(uint32_t *)(APP_START_ADDR + 4);

    printf("\r\nJumping to the application\r\n");
    /* Rebase the Stack Pointer */
    __set_MSP(*(uint32_t*)APP_START_ADDR);  
     /* Rebase the vector table base address */
    SCB->VTOR = ((uint32_t) APP_START_ADDR & SCB_VTOR_TBLOFF_Msk); 
    /* Jump to application Reset Handler in the application */
    asm("bx %0"::"r"(app_start_addr));
}

// *****************************************************************************
// *****************************************************************************
// Section: Application Initialization and State Machine Functions
// *****************************************************************************
// *****************************************************************************

/*******************************************************************************
  Function:
    void APP_Initialize ( void )

  Remarks:
    See prototype in app.h.
 */

void APP_Initialize ( void )
{
    /* Place the App state machine in its initial state. */
    appData.state = APP_STATE_INIT;



    /* TODO: Initialize your application's state machine and other
     * parameters.
     */
}


/******************************************************************************
  Function:
    void APP_Tasks ( void )

  Remarks:
    See prototype in app.h.
 */

void APP_Tasks ( void )
{
    ATCA_STATUS status;

    /* Check the application's current state. */
    switch ( appData.state )
    {
        /* Application's initial state. */
        case APP_STATE_INIT:
            status = atcab_init(&atecc608_0_init_data);
            appData.state = (status == ATCA_SUCCESS) ? APP_STATE_CALC_DIGEST_GET_SIGN: APP_STATE_FINISH;
            break;

        case APP_STATE_CALC_DIGEST_GET_SIGN:
            status = calc_digest_and_get_signature();
            appData.state = (status == ATCA_SUCCESS) ? APP_STATE_VALIDATE_FIRM_BASED_ON_DIGEST: APP_STATE_FINISH;
            break;

        case APP_STATE_VALIDATE_FIRM_BASED_ON_DIGEST:
            status = validate_firmware_based_on_stored_digest();
            appData.state = (status == ATCA_SUCCESS) ? APP_STATE_JUMP_TO_APPLICATION: APP_STATE_VALIDATE_FIRM_AND_UPDATE_DIGEST;
            break;

        case APP_STATE_VALIDATE_FIRM_AND_UPDATE_DIGEST:
            status = validate_firmware_and_update_digest();
            appData.state = (status == ATCA_SUCCESS) ? APP_STATE_JUMP_TO_APPLICATION: APP_STATE_FINISH;
            break;

        case APP_STATE_JUMP_TO_APPLICATION:
            jump_to_application();
            break;

        case APP_STATE_FINISH:
            break;

        /* TODO: implement your application state machine.*/


        /* The default state should never be executed. */
        default:
        {
            /* TODO: Handle error in application's state machine. */
            break;
        }
    }
}


/*******************************************************************************
 End of File
 */
